#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <GL/glew.h>	/* glew.h prin apo opoiodipote GL i glut header file */
#include <GL/glut.h>
#include "image_rgba.h"

#ifndef M_PI
#define M_PI	3.14159265358
#endif

int init();
void update(float time_sec);
void move_player(float dx, float dy);
void display();
void draw_floor();
void draw_teapot();
void idle();
void reshape(int x, int y);
void keyboard(unsigned char key, int x, int y);
void keyboard_up(unsigned char key, int x, int y);
void mouse(int button, int state, int x, int y);
void motion(int x, int y);
unsigned int load_texture_rgba(const char *fname);

/* --- global variables --- */
int screen_width, screen_height;
bool keystate[256];

/* player position and orientation */
bool first_person;
float player_pos_x, player_pos_y = 1.5, player_pos_z = 8;
float player_theta, player_phi;

float walk_speed = 5.0;		// taxytita perpatimatos
float look_speed = 140.0;	// taxytita peristrofis


unsigned int floor_tex;


int main(int argc, char **argv)
{
	glutInit(&argc, argv);
	glutInitWindowSize(800, 600);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutCreateWindow("OpenGL ergasia");

	glutDisplayFunc(display);
	glutIdleFunc(idle);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);
	glutKeyboardUpFunc(keyboard_up);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);

	if(init() == -1) {
		return 1;
	}

	glutMainLoop();
	return 0;
}


/* --- initialization, ekteleite stin arxi apo tin main, epistrefei -1 an yparksei provlima --- */
int init()
{
	/* panta otan xrisimopoioume GLEW, prota kaloume tin glewInit */
	glewInit();

	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	floor_tex = load_texture_rgba("data/floor.tga");
	if(!floor_tex) {
		return -1;
	}

	return 0;
}

void update(float time_sec)
{
	static float prev_update;
	float dt = time_sec - prev_update;
	prev_update = time_sec;

	float step = walk_speed * dt;

	// handle input
	if(keystate['d']) {
		move_player(step, 0);
	}
	if(keystate['a']) {
		move_player(-step, 0);
	}
	if(keystate['w']) {
		move_player(0, step);
	}
	if(keystate['s']) {
		move_player(0, -step);
	}
}

/* metakinei ton paixti kata dx monades aristera, kai dy monades mprosta
 * sxetika me tin kateythinsi pou koitaei
 */
void move_player(float dx, float dy)
{
	float angle = M_PI * player_theta / 180.0;
	player_pos_x += cos(angle) * dx - sin(angle) * dy;
	player_pos_z -= sin(angle) * dx + cos(angle) * dy;
}

void display()
{
	unsigned int msec = glutGet(GLUT_ELAPSED_TIME);

	// i update xeirizetai tin kinisi tou paixti kai ananeonei tin katastasi tou kosmou
	update(msec / 1000.0);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/* --- setup view (camera) matrix --- */
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glRotatef(-player_phi, 1, 0, 0);
	glRotatef(-player_theta, 0, 1, 0);
	glTranslatef(-player_pos_x, -player_pos_y, -player_pos_z);
	/* --- end of view matrix ---*/

	/* --- set light position --- */
	float lpos[] = {-10, 20, 10, 1};
	glLightfv(GL_LIGHT0, GL_POSITION, lpos);

	/* --- render all objects, thetontas prota to world matrix tous --- */
	draw_floor();

	glPushMatrix();
	glTranslatef(0, 0.75, 0);
	draw_teapot();
	glPopMatrix();

	/* --- sto telos glutSwapBuffers gia na emfanistei to apotelesma --- */
	glutSwapBuffers();
	assert(glGetError() == GL_NO_ERROR);
}

void draw_floor()
{
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, floor_tex);

	glBegin(GL_QUADS);
	glNormal3f(0, 1, 0);
	glTexCoord2f(0, 0);
	glVertex3f(-20, 0, 20);
	glTexCoord2f(8, 0);
	glVertex3f(20, 0, 20);
	glTexCoord2f(8, 8);
	glVertex3f(20, 0, -20);
	glTexCoord2f(0, 8);
	glVertex3f(-20, 0, -20);
	glEnd();

	glDisable(GL_TEXTURE_2D);
}

void draw_teapot()
{
	glFrontFace(GL_CW);
	glutSolidTeapot(1.0);
	glFrontFace(GL_CCW);
}

void idle()
{
	glutPostRedisplay();
}

void reshape(int x, int y)
{
	glViewport(0, 0, x, y);
	screen_width = x;
	screen_height = y;

	/* --- set the projection matrix --- */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(50.0, (float)x / (float)y, 0.5, 500.0);
}

void keyboard(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
	case 'q':
		exit(0);

	case '`':
		first_person = !first_person;
		if(first_person) {
			glutSetCursor(GLUT_CURSOR_NONE);
			glutPassiveMotionFunc(motion);
			glutWarpPointer(screen_width / 2, screen_height / 2);
		} else {
			glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
			glutPassiveMotionFunc(0);
		}
		break;

	case 'f':
		{
			static bool fullscreen;
			static int orig_width, orig_height;

			fullscreen = !fullscreen;
			if(fullscreen) {
				orig_width = screen_width;
				orig_height = screen_height;
				glutFullScreen();
			} else {
				glutReshapeWindow(orig_width, orig_height);
			}
		}
		break;
	}

	keystate[key] = true;
}

void keyboard_up(unsigned char key, int x, int y)
{
	keystate[key] = false;
}


int bnstate[16];
int prev_x, prev_y;

void mouse(int button, int state, int x, int y)
{
	/* kratame tin katastasi ton pliktron sto array bnstate */
	int bidx = button - GLUT_LEFT_BUTTON;
	if(state == GLUT_DOWN) {
		bnstate[bidx] = 1;

		/* thetoume to prev_x kai prev_y sto simio tou pontikiou otan patithike kapoio pliktro */
		prev_x = x;
		prev_y = y;
	} else {
		bnstate[bidx] = 0;
	}
}

void motion(int x, int y)
{
	float dx = (float)(x - prev_x) / (float)screen_width;
	float dy = (float)(y - prev_y) / (float)screen_height;
	prev_x = x;
	prev_y = y;

	if(first_person) {
		int center_x = screen_width / 2;
		int center_y = screen_height / 2;

		if(abs(x - center_x) == 0 && abs(y - center_y) == 0) {
			return;
		}

		dx = (float)(x - center_x) / (float)screen_width;
		dy = (float)(y - center_y) / (float)screen_height;
		glutWarpPointer(center_x, center_y);
	}

	if(first_person || bnstate[0]) {
		player_theta -= dx * look_speed;
		player_phi -= dy * look_speed;

		if(player_phi < -90) player_phi = -90;
		if(player_phi > 90) player_phi = 90;
	}
}

unsigned int load_texture_rgba(const char *fname)
{
	struct image_rgba *img = load_image_rgba(fname);
	if(!img) {
		return 0;
	}


	unsigned int tex;
	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);

	// enable mip-mapping if mipmap generation is available
	if(GLEW_SGIS_generate_mipmap) {
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, 1);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	} else {
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	}

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, img->width, img->height, 0, GL_RGBA, GL_UNSIGNED_BYTE, img->pixels);
	free_image_rgba(img);

	return tex;
}
